import pytest

from rumydata import table, Layout, field
from rumydata.rules import row as rr, header as hr
from tests.utils import file_row_harness


def test_row_good(basic):
    row = ['1', '2', '2020-01-01', 'X']
    assert not Layout(basic).check_row(row)
    assert not file_row_harness(row, basic)


def test_row_choice(basic):
    fields = {'c1': field.Choice(['x'], nullable=True)}
    lay = Layout(fields)
    assert not lay.check_row(['x'])
    assert not lay.check_row([''])
    assert not file_row_harness(['x'], fields)
    assert not file_row_harness([''], fields)


@pytest.mark.parametrize('value,err', [
    ([1, 2, 3, 4, 5], rr.RowLengthLTE),
    ([1, 2, 3], rr.RowLengthGTE)
])
def test_row_bad(basic, value, err):
    assert table.Layout(basic)._has_error(value, err.rule_exception(), rule_type=rr.Rule)
    with pytest.raises(AssertionError):
        file_row_harness(value, basic)


def test_header_good(basic):
    assert not table.Layout(basic).check_header(['col1', 'col2', 'col3', 'col4'])

@pytest.mark.parametrize('value,err', [
    (['col1', 'col2'], hr.NoMissing),
    (['col1', 'col2', 'col2'], hr.NoDuplicate),
    (['col1', 'col2', 'col5'], hr.NoExtra)
])
def test_header_bad(basic, value, err):
    assert table.Layout(basic)._has_error(value, err.rule_exception(), rule_type=hr.Rule)


def test_header_skip(basic):
    assert not table.Layout(basic, skip_header=True). \
        check_header(['col1', 'col2', 'col4'])


@pytest.mark.parametrize('row_length,row,expected', [
    (1, ['1', '2'], False),
    (1, ['1'], True)
])
def test_row_length_lte(row_length, row, expected):
    r = rr.RowLengthLTE(row_length)
    assert r._evaluator()(*r._prepare(row)) is expected


@pytest.mark.parametrize('row_length,row,expected', [
    (3, ['1', '2'], False),
    (2, ['1', '2'], True),
    (1, ['1', '2'], True)
])
def test_row_length_gte(row_length, row, expected):
    r = rr.RowLengthGTE(row_length)
    assert r._evaluator()(*r._prepare(row)) is expected



import csv
import tempfile
from pathlib import Path
from typing import List, Union, Tuple
from unittest.mock import DEFAULT

import openpyxl

from rumydata import Layout, CsvFile, ExcelFile
from rumydata.field import Field


def mock_no_module(module: str):
    """ force exception on specified module import. For use with mocker """

    def func(*args):
        if args[0] == module:
            raise ModuleNotFoundError
        else:
            return DEFAULT

    return func


def file_row_harness(row: List[Union[str, int]], layout: dict):
    """ Write row to file for testing in ingest """
    lay = Layout(layout, no_header=True)

    with tempfile.TemporaryDirectory() as d:
        csv_p = Path(d, 'file_test.csv')

        with csv_p.open('w', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(row)

        xl_p = Path(d, 'excel_test.xlsx')
        wb = openpyxl.Workbook()
        sheet = wb['Sheet']
        for ix, value in enumerate(row, start=1):
            sheet.cell(row=1, column=ix).value = value
        wb.save(filename=xl_p)

        to_check = [
            ('CsvFile', CsvFile(lay), csv_p),
            ('ExcelFile', ExcelFile(lay), xl_p)
        ]
        aes = {}
        for nm, obj, p in to_check:
            try:
                assert not obj.check(p)
            except AssertionError as e:
                aes[nm] = e
        new_line = '\n'
        assert not aes, f'Write test failed for:\n {new_line.join([f"{k}:{v}" for k, v in aes.items()])}'


def file_cell_harness(value: str, field: Field):
    """ Wrapper to convert cell to row for harness check """
    file_row_harness(row=[value], layout={'c1': field})



"""
field/column/element objects

This submodule contains the various field type objects that are used to define
the fields (i.e. columns, elements) that make up a tabular data set. These field
objects can be extended by adding additional rules in the rules keyword of the
fields constructor. Additional modification, including the definition of
entirely new field types can be achieved by subclassing the Field base class.

All field subclasses pass their keyword arguments to the base Field class. See
documentation for the base Field for a complete listing of optional keyword
arguments, including rules extension.

See the rumydata.rules submodule to learn more about the use of rules to extend
field class behavior.
"""
from typing import Union, Tuple, Dict, List

from rumydata import exception as ex
from rumydata._base import _BaseSubject
from rumydata.rules import cell as clr, column as cr

__all__ = ['Text', 'Date', 'Currency', 'Digit', 'Integer', 'Choice', 'Ignore', 'Empty']


class Field(_BaseSubject):
    """
    Base Field class

    This class provides the framework for the definition of field types in this
    package.

    :param nullable: a boolean indicator which controls whether the field
        can be null (blank). Defaults to False, which will cause errors to
        be raised when checking empty strings.
    :param rules: a list of rules to apply to this field during checks.
    :param strip: (optional) apply pre-processing to values in the field which
        applies the str.strip method, removing leading and trailing whitespaces,
        prior to checking rules.
    """

    def __init__(self, nullable=False, rules: list = None, **kwargs):
        self.strip = kwargs.pop('strip', None)
        self._ignore_if = None
        super().__init__(rules, **kwargs)
        self.nullable = nullable

        if not self.nullable:
            self.rules.append(clr.NotNull())

    def check_cell(self, value: Union[str, Tuple[str, Dict]], **kwargs):
        """
        Cell Rule assertion

        Perform an assertion of the provided cell value against the rules
        defined for the field. If the value fails the check for any of the
        rules, the assertion will raise a detailed exception message.

        :param value: a cell value, either a string or a tuple of a string
            and a dictionary, to be checked.
        """

        errors = self._check(value, rule_type=clr.Rule, **kwargs)
        assert not errors, str(errors)

    def check_column(self, values: List[str], **kwargs):
        """
        Column Rule assertion

        Perform an assertion of the provided column values against the rules
        defined for the field. If the values fail the check for any of the
        rules, the assertion will raise a detailed exception message.

        :param values: a list of values contained in the column.
        """

        errors = self._check(values, rule_type=cr.Rule, **kwargs)
        assert not errors, str(errors)

    def _check_for_nullable_rules(self):
        return any([x for x in self.rules if isinstance(x, clr.NotNullIfCompare)])

    def _check_nullable_rule_results(self, data):
        return all([x._null_ok(data) for x in self.rules if isinstance(x, clr.NotNullIfCompare)])

    def _check(self, data, cix=-1, rule_type=None, **kwargs) -> Union[ex.CellError, ex.ColumnError, None]:
        """
        Check data against field rules of specified rule type

        This is the core method used to evaluate data against a subject defined
        by this class.

        :param data: an object which conforms to the `prepare` method of the
            class specified in the rule_type parameter
        :param cix: column index used for error reporting.
        :param rule_type: a Rule class belonging to one of the submodules in the
            rules module (e.g. rumydata.rules.cell.Rule). This controls the
            types of rules that the provided data will be checked against.
        :return: a list of any errors that were raised while checking the data.
        """

        # if data is nullable and value is empty, skip all checks
        empty = data[0] == '' if isinstance(data, tuple) else data == ''
        if self.nullable and rule_type == clr.Rule and empty:
            pass
        elif empty and (self._check_nullable_rule_results(data) if self._check_for_nullable_rules() else False):
            pass
        else:
            e = super()._check(data, rule_type=rule_type, strip=self.strip)
            if e:
                if rule_type == cr.Rule:
                    return ex.ColumnError(cix, errors=e, **kwargs)
                else:
                    if self.include_all_errors:
                        if self.custom_error_msg:
                            e.insert(0, ex.CustomError(self.custom_error_msg))
                        return ex.CellError(cix, errors=e, **kwargs)
                    else:
                        if self.custom_error_msg:
                            return ex.CellError(cix, errors=[ex.CustomError(self.custom_error_msg)], **kwargs)
                        else:
                            return ex.CellError(cix, **kwargs)

    def _comparison_columns(self) -> set:
        """
        Comparison fields report

        A method to report the columns that will need to be compared while
        checking rules.

        :return: a set of the columns that will need to be compared.
        """
        compares = set()
        for r in self.rules:
            if issubclass(type(r), clr.ColumnComparisonRule):
                compares.add(r.compare_to) if isinstance(r.compare_to, str) else [compares.add(x) for x in r.compare_to]
        return compares

    def _has_rule_type(self, rule_type):
        return any([issubclass(type(r), rule_type) for r in self.rules])

    def _digest(self):
        dig = super()._digest()
        if self.nullable:
            dig.append('Nullable')
        return dig


class Ignore(Field):
    """
    Ignore the values found in this field. This will cause other dependencies
    (like row checks) to treat any values in this field as if they were empty.
    While the check method still exists, this class overwrites those inherited
    methods to intentionally have no effect.
    """

    # noinspection PyMissingConstructor
    def __init__(self):
        self.nullable = True
        self.rules = []
        self.descriptors = {}

    def _check(self, *args, **kwargs):
        pass


class Empty(Field):
    """
    Empty Field

    A field which must be entirely empty/blank/null. This differs from the Ignore
    field in that it will raise errors if any values are found.
    """

    def __init__(self, **kwargs):
        super().__init__(nullable=True, **kwargs)
        self.rules.append(clr.MaxChar(0))


class Text(Field):
    """
    Text Field

    A field made up entirely of text. This is one of the least restrictive field
    types, enforcing only a maximum and (optional) minimum length.

    :param max_length: the maximum number of allowable characters
    :param min_length: (optional) the minimum number of allowable characters
    """

    _default_args = (1,)

    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'String'
        self.descriptors['Max Length'] = f'{str(max_length)} characters'

        self.rules.append(clr.MaxChar(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(min_length)} characters'
            self.rules.append(clr.MinChar(min_length))


class Date(Field):
    """
    Date field

    A date value formatted in ISO-8601 (YYYY-MM-DD). Rules can be applied to
    restrict minimum and maximum date. If your date string includes, or might
    include an empty timestamp (00:00:00), the truncate_time parameter can be
    used to remove this empty timestamp prior to validation, preventing it from
    causing failure.

    :param min_date: the minimum date value allowed
    :param max_date: the maximum date value allowed
    """

    def __init__(self, min_date: str = None, max_date: str = None, **kwargs):
        rule_kwargs = {
            'truncate_time': kwargs.pop('truncate_time', False)
        }
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Date'
        self.descriptors['Format'] = 'YYYY-MM-DD'

        self.rules.append(clr.CanBeDateIso(**rule_kwargs))
        if max_date:
            self.descriptors['Max Date'] = f'{max_date}'
            self.rules.append(clr.DateLTE(max_date, **rule_kwargs))

        if min_date:
            self.descriptors['Min Date'] = f'{min_date}'
            self.rules.append(clr.DateGTE(min_date, **rule_kwargs))


class Currency(Field):
    """
    Currency field

    A numeric field which represents a currency value. Defaults to a maximum
    of two decimal points.

    :param significant_digits: an integer which specifies the maximum number of
        decimals that are allowed in the value.
    """

    _default_args = (5,)

    def __init__(self, significant_digits: int, precision: int = 2, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = \
            f'{"9" * (significant_digits - precision)}.{"9" * precision}'
        self.descriptors['Max Length'] = f'{str(significant_digits)} digits'
        self.rules.append(clr.MaxDigit(significant_digits))
        self.rules.append(clr.NumericDecimals(precision))


class Digit(Field):
    """
    Digit field

    A value made up entirely of digits (numbers). The main difference between
    this field and an Integer, is that this field will allow leading zeroes
    (e.g 0123), while an Integer would not.

    :param max_length: the maximum number of digits
    :param min_length: (optional) the minimum number of digits
    """

    _default_args = (1,)

    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = f'{"0" * max_length}'
        self.descriptors['Max Length'] = f'{str(max_length)} digits'

        self.rules.append(clr.OnlyNumbers())
        self.rules.append(clr.MaxChar(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(min_length)} digits'
            self.rules.append(clr.MinChar(min_length))


class Integer(Field):
    """
    Integer field

    A value made up entirely of digits (numbers). A whole number.

    :param max_length: the maximum number of digits
    :param min_length: (optional) the minimum number of digits
    """

    _default_args = (1,)

    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = f'{"9" * max_length}'
        self.descriptors['Max Length'] = f'{str(max_length)} digits'

        self.rules.append(clr.CanBeInteger())
        self.rules.append(clr.NoLeadingZero())
        self.rules.append(clr.MaxDigit(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(max_length)} digits'
            self.rules.append(clr.MinDigit(min_length))


class Choice(Field):
    """
    Choice field

    A field with one of a pre-defined set of values.

    :param valid_values: a list of values which are considered valid
    :param case_insensitive: a boolean control which defaults to False, making
        the check of valid values case sensitive.
    """

    _default_args = (['x'],)

    def __init__(self, valid_values: list, case_insensitive=False, **kwargs):
        super().__init__(**kwargs)
        self.descriptors['Type'] = 'Choice'
        self.descriptors['Choices'] = ','.join(valid_values)
        self.rules.append(
            clr.Choice(valid_values, case_insensitive=case_insensitive)
        )



from pathlib import Path
from uuid import uuid4

import pytest

from rumydata.field import Integer
from rumydata.menu import menu
from rumydata.table import Layout
from tests.utils import mock_no_module


# noinspection DuplicatedCode
@pytest.mark.parametrize('choice', ['0', 'View Documentation'])
@pytest.mark.parametrize('no_md,ext', [(False, 'html'), (True, 'md')])
def test_view_documentation(choice, ext, no_md, mocker):
    """
    View Documentation option opens a browser with generated documentation. When
    the markdown module is available, this is shown in HTML, otherwise displayed
    as a raw .md text file.
    """
    if no_md:
        mocker.patch(
            'builtins.__import__', wraps=__import__, side_effect=mock_no_module('markdown')
        )

    mocker.patch('webbrowser.open')
    mocker.patch('builtins.print')
    mocker.patch('builtins.input', return_value=choice)

    expected = dict(extension=ext, output='open')
    ret = menu(Layout({'x': Integer(1)}))
    for k, v in expected.items():
        assert ret[k] == v


# noinspection DuplicatedCode
@pytest.mark.parametrize('choice', ['2', 'View Validation'])
@pytest.mark.parametrize('no_md,ext', [(False, 'html'), (True, 'md')])
@pytest.mark.parametrize('valid_file', [False, True])
def test_view_validation(choice, ext, no_md, valid_file, tmpdir, mocker):
    """
    View Validation option opens a browser with generated documentation. When
    the markdown module is available, this is shown in HTML, otherwise displayed
    as a raw .md text file.
    """
    if no_md:
        mocker.patch(
            'builtins.__import__', wraps=__import__, side_effect=mock_no_module('markdown')
        )
    if valid_file:
        p = Path(tmpdir, 'file.csv')
        p.write_text('x\n1\n')
        choice = (choice, p.as_posix())
        print(choice)
    else:
        choice = (choice, 'file.csv')

    mocker.patch('webbrowser.open')
    mocker.patch('builtins.print')
    mocker.patch('builtins.input', side_effect=choice)

    expected = dict(extension=ext, output='open')
    ret = menu(Layout({'x': Integer(1)}))
    for k, v in expected.items():
        assert ret[k] == v


@pytest.mark.parametrize('no_md', [False, True])
@pytest.mark.parametrize('choice,expected', [
    (['1', '0', '0'], dict(extension='md')),
    (['Generate Documentation', 'markdown', 'print'], dict(extension='md')),
    (['Generate Documentation', 'html', 'print'], dict(extension='html')),
    (['3', '0', 'x', '0'], dict(extension='md')),
    (['Generate Validation', 'markdown', 'x', 'print'], dict(extension='md')),
    (['Generate Validation', 'html', 'x', 'print'], dict(extension='html')),
])
def test_generate_print(choice, expected: dict, no_md, mocker):
    if no_md:
        mocker.patch(
            'builtins.__import__', wraps=__import__, side_effect=mock_no_module('markdown')
        )
        expected['extension'] = 'md'

    mocker.patch('webbrowser.open')
    mocker.patch('builtins.print')
    mocker.patch('builtins.input', side_effect=choice)

    expected['output'] = 'print'

    ret = menu(Layout({'x': Integer(1)}))
    for k, v in expected.items():
        assert ret[k] == v


@pytest.mark.parametrize('no_md', [False, True])
@pytest.mark.parametrize('choice,expected', [
    (['1', '0', '2'], dict(extension='md')),
    (['Generate Documentation', 'markdown', 'save'], dict(extension='md')),
    (['Generate Documentation', 'html', 'save'], dict(extension='html')),
    (['3', '0', 'x', '2'], dict(extension='md')),
    (['Generate Validation', 'markdown', 'x', 'save'], dict(extension='md')),
    (['Generate Validation', 'html', 'x', 'save'], dict(extension='html')),
])
def test_generate_save(choice, expected: dict, no_md, tmpdir, mocker):
    if no_md:
        mocker.patch(
            'builtins.__import__', wraps=__import__, side_effect=mock_no_module('markdown')
        )
        expected['extension'] = 'md'

    mocker.patch('webbrowser.open')
    mocker.patch('builtins.print')

    p = Path(tmpdir, uuid4().hex[:5])
    mocker.patch('builtins.input', side_effect=choice + [p.as_posix()])

    expected['output'] = 'save'

    ret = menu(Layout({'x': Integer(1)}))
    for k, v in expected.items():
        assert ret[k] == v



"""
cell validation rules

These rules make up the heart of what most users of the rumydata package will
be interested in when attempting to extend the out-of-the box behavior. These
rules are generally applied to a single value, in the case of the Rule class,
but can also be used to compare the value in a cell to another value in the same
row, in the case of the ColumnComparisonRule class.

These rules are intended to be used by adding them directly to rules argument in
the constructor of the classes in the field submodule.
"""
import re
from datetime import datetime
from typing import Union, Tuple, Dict, List

from rumydata._base import _BaseRule

__all__ = [
    'NotNull', 'ExactChar', 'MinChar', 'MaxChar', 'AsciiChar', 'NonTrim', 'Choice',
    'MinDigit', 'MaxDigit', 'OnlyNumbers', 'NoLeadingZero', 'CanBeFloat',
    'CanBeInteger', 'NumericDecimals', 'LengthComparison', 'LengthGT',
    'LengthGTE', 'LengthET', 'LengthLTE', 'LengthLT', 'NumericComparison',
    'NumericGT', 'NumericGTE', 'NumericET', 'NumericLTE', 'NumericLT',
    'DateRule', 'CanBeDateIso', 'DateGT', 'DateGTE', 'DateET', 'DateLTE',
    'DateLT', 'GreaterThanColumn', 'NotNullIfCompare', 'GreaterThanOrEqualColumn',
    'OtherMustExist', 'OtherCantExist', 'LessThanColumn', 'LessThanOrEqualColumn',
    'NotNullIfOtherEquals',
    'make_static_cell_rule'
]


class Rule(_BaseRule):
    """ Cell Rule """

    @staticmethod
    def _pre_process(data: Union[str, Tuple[str, Dict]], **kwargs) -> Union[str, Tuple[str, Dict]]:
        d1 = data if isinstance(data, str) else data[0]
        d2 = data[1] if isinstance(data, tuple) else {}

        if kwargs.get('strip'):
            d1 = d1.strip()
            d2 = {k: v.strip() for k, v in d2.items()}

        if d2:
            return d1, d2
        else:
            return d1

    def _prepare(self, data: Union[str, Tuple[str, Dict]]) -> tuple:
        if isinstance(data, str):
            return data,
        else:
            return data[0],


def make_static_cell_rule(func, assertion) -> Rule:
    """
    Static cell rule factory

    Return a factory generated Rule class. The function used by the rule must
    directly evaluate a single positional argument (i.e. x, but not x and y).
    Because the Rule cannot be passed a value on initialization, neither the
    evaluator or explain methods in the return class can be dynamic.

    :param func: a function which takes a single positional argument
    :param assertion: a string describing the condition which must be met in
        order for the function to return True
    :return: a rumydata.rules.cell.Rule
    """

    class FactoryRule(Rule):

        def _evaluator(self):
            return func

        def _explain(self) -> str:
            return assertion

    return FactoryRule()


class NotNull(Rule):
    """ Cell not null Rule """

    def _evaluator(self):
        return lambda x: x != ''

    def _explain(self) -> str:
        return 'cannot be empty/blank'


class ExactChar(Rule):
    """ Cell exact character length Rule """

    _default_args = (1,)

    def __init__(self, exact_length):
        super().__init__()
        self.exact_length = exact_length

    def _evaluator(self):
        return lambda x: len(x) == self.exact_length

    def _explain(self) -> str:
        return f'must be exactly {str(self.exact_length)} characters'


class MinChar(Rule):
    """ Cell minimum character length Rule """

    _default_args = (1,)

    def __init__(self, min_length):
        super().__init__()
        self.min_length = min_length

    def _evaluator(self):
        return lambda x: len(x) >= self.min_length

    def _explain(self) -> str:
        return f'must be at least {str(self.min_length)} characters'


class MaxChar(Rule):
    """ Cell maximum character length Rule """

    _default_args = (1,)

    def __init__(self, max_length):
        super().__init__()
        self.max_length = max_length

    def _evaluator(self):
        return lambda x: len(x) <= self.max_length

    def _explain(self) -> str:
        return f'must be no more than {str(self.max_length)} characters'


class AsciiChar(Rule):
    """ Cell contains only ASCII character Rule """

    def _evaluator(self):
        return lambda x: all(ord(c) < 128 for c in x)

    def _explain(self) -> str:
        return 'must have only ASCII characters'


class NonTrim(Rule):
    """ Cell does not have whitespace characters at beginning or end """

    def _evaluator(self):
        return lambda x: x == str(x).strip()

    def _explain(self) -> str:
        return 'Value has trailing or leading whitespace'


class Choice(Rule):
    """ Cell choice Rule """

    _default_args = (['x'],)

    def __init__(self, choices: List[str], case_insensitive=False):
        super().__init__()
        self.choices = choices
        self.case_insensitive = case_insensitive
        self.eval_choices = [x.lower() for x in choices] if case_insensitive else choices

    def _prepare(self, data: Union[List[str], Tuple[str, Dict]]) -> tuple:
        if self.case_insensitive:
            if isinstance(data, str):
                data = data.lower(),
            else:
                data = data[0].lower(), data[1]

        return super()._prepare(data)

    def _evaluator(self):
        return lambda x: x in self.eval_choices

    def _explain(self) -> str:
        return (
            f'must be one of {self.choices}'
            f'{" (case insensitive)" if self.case_insensitive else " (case sensitive)"}'
        )


class MinDigit(Rule):
    """
    Cell minimum digit character Rule

    Check that count of characters, after removing all non-digits, meets or
    exceeds the specified minimum. Used to evaluate length of significant digits
    in numeric strings that might contain formatting.
    """
    _default_args = (1,)

    def __init__(self, min_length):
        super().__init__()
        self.min_length = min_length

    def _evaluator(self):
        return lambda x: len(re.sub(r'[^\d]', '', x)) >= self.min_length

    def _explain(self) -> str:
        return f'must have at least {str(self.min_length)} digit characters'


class MaxDigit(Rule):
    """
    Cell maximum digit character Rule

    Check that count of characters, after removing all non-digits, is less than
    or equal to the specified minimum. Used to evaluate length of significant
    digits in numeric strings that might contain formatting.
    """
    _default_args = (1,)

    def __init__(self, max_length):
        super().__init__()
        self.max_length = max_length

    def _evaluator(self):
        return lambda x: len(re.sub(r'[^\d]', '', x)) <= self.max_length

    def _explain(self) -> str:
        return f'must have no more than {self.max_length} digit characters'


class OnlyNumbers(Rule):
    """ Cell only digit characters Rule """

    def _evaluator(self):
        return lambda x: bool(re.fullmatch(r'\d+', x))

    def _explain(self) -> str:
        return 'must only contain characters 0-9'


class NoLeadingZero(Rule):
    """
    Cell no leading zero digit Rule

    Ensure that there is no leading zero after removing all non-digit characters.
    A lone zero (0) will not raise an error.
    """

    def _evaluator(self):
        return lambda x: bool(re.fullmatch(r'(0|([1-9]\d*))', re.sub(r'[^\d]', '', x)))

    def _explain(self) -> str:
        return 'cannot have a leading zero digit'


class CanBeFloat(Rule):
    """ Cell can be float Rule """

    def _evaluator(self):
        def fun(x):
            try:
                return isinstance(float(x), float)
            except ValueError:
                return False

        return fun

    def _explain(self) -> str:
        return 'can be coerced into a float value'


class CanBeInteger(Rule):
    """ Cell can be integer Rule """

    def _evaluator(self):
        def fun(x):
            try:
                return isinstance(int(x), int)
            except ValueError:
                return False

        return fun

    def _explain(self) -> str:
        return 'can be coerced into an integer value'


class NumericDecimals(Rule):
    """ Cell has maximum decimals Rule """

    def __init__(self, max_decimals=2):
        super().__init__()
        self.decimals = max_decimals

    def _evaluator(self):
        pat = re.compile(r'-?\d+(\.\d{1,' + str(self.decimals) + '})?')
        return lambda x: bool(pat.fullmatch(x))

    def _explain(self) -> str:
        return f'cannot have more than {self.decimals} digits after the decimal point'


class LengthComparison(Rule):
    """ Base length comparison Rule """

    comparison_language = 'N/A'
    _default_args = ('x',)

    def __init__(self, comparison_value):
        super().__init__()
        self.comparison_value = comparison_value

    def _explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class LengthGT(LengthComparison):
    """ Length greater than comparison Rule """

    comparison_language = 'greater than'

    def _evaluator(self):
        return lambda x: len(x) > self.comparison_value


class LengthGTE(LengthComparison):
    """ Length greater than or equal to comparison Rule """

    comparison_language = 'greater than or equal to'

    def _evaluator(self):
        return lambda x: len(x) >= self.comparison_value


class LengthET(LengthComparison):
    """ Length equal to comparison Rule """

    comparison_language = 'equal to'

    def _evaluator(self):
        return lambda x: len(x) == self.comparison_value


class LengthLTE(LengthComparison):
    """ Length less than or equal to comparison Rule """

    comparison_language = 'less than or equal to'

    def _evaluator(self):
        return lambda x: len(x) <= self.comparison_value


class LengthLT(LengthComparison):
    """ Length less than comparison Rule """

    comparison_language = 'less than'

    def _evaluator(self):
        return lambda x: len(x) < self.comparison_value


class NumericComparison(Rule):
    """
    Numeric length comparison base Rule

    Base float value comparison class. Requires that the value can be coerced
    to a float value.
    """

    comparison_language = 'N/A'
    _default_args = ('x',)

    def __init__(self, comparison_value):
        super().__init__()
        self.comparison_value = comparison_value

    def _explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class NumericGT(NumericComparison):
    """ Numeric greater than comparison Rule """

    comparison_language = 'greater than'

    def _evaluator(self):
        return lambda x: float(x) > self.comparison_value


class NumericGTE(NumericComparison):
    """ Numeric greater than or equal to comparison Rule """

    comparison_language = 'greater than or equal to'

    def _evaluator(self):
        return lambda x: float(x) >= self.comparison_value


class NumericET(NumericComparison):
    """ Numeric equal to comparison Rule """

    comparison_language = 'equal to'

    def _evaluator(self):
        return lambda x: float(x) == self.comparison_value


class NumericLTE(NumericComparison):
    """ Numeric less than or equal to comparison Rule """

    comparison_language = 'less than or equal to'

    def _evaluator(self):
        return lambda x: float(x) <= self.comparison_value


class NumericLT(NumericComparison):
    """ Numeric less than comparison Rule """
    comparison_language = 'less than'

    def _evaluator(self):
        return lambda x: float(x) < self.comparison_value


class DateRule(Rule):
    """ Base date Rule """

    def __init__(self, **kwargs):
        self.truncate_time = kwargs.pop('truncate_time', False)
        super().__init__()

    def _prepare(self, data: Union[str, Tuple[str, Dict]]) -> tuple:
        if self.truncate_time:
            no_time = ' 00:00:00'
            if isinstance(data, str) and data.endswith(no_time):
                data = data[:-len(no_time)]
            elif data[0].endswith(no_time):
                data = data[0][:-len(no_time)], data[1]

        return super()._prepare(data)


class CanBeDateIso(DateRule):
    """ Can be ISO-8601 date Rule """

    def _evaluator(self):
        def func(x):
            try:
                return isinstance(datetime.strptime(x, '%Y-%m-%d'), datetime)
            except ValueError:
                return False

        return func

    def _explain(self) -> str:
        return 'can be coerced into a ISO-8601 date'


class DateComparisonRule(DateRule):
    """
    Base date comparison Rule

    Base date value comparison class. Requires that the value can be coerced
    to a date using the specified format for the field.
    """

    comparison_language = 'N/A'
    _default_args = ('2020-01-01',)

    def __init__(self, comparison_value, date_format='%Y-%m-%d', **kwargs):
        self.date_format = date_format
        self.comparison_value = datetime.strptime(comparison_value, date_format)
        super().__init__(**kwargs)

    def _explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class DateGT(DateComparisonRule):
    """ Date greater than comparison Rule """

    comparison_language = 'greater than'

    def _evaluator(self):
        def func(x):
            try:
                return datetime.strptime(x, self.date_format) > self.comparison_value
            except ValueError:
                return False

        return func


class DateGTE(DateComparisonRule):
    """ Date greater than or equal to comparison """

    comparison_language = 'greater than or equal to'

    def _evaluator(self):
        def func(x):
            try:
                return datetime.strptime(x, self.date_format) >= self.comparison_value
            except ValueError:
                return False

        return func


class DateET(DateComparisonRule):
    """ Date equal to comparison Rule """

    comparison_language = 'equal to'

    def _evaluator(self):
        def func(x):
            try:
                return datetime.strptime(x, self.date_format) == self.comparison_value
            except ValueError:
                return False

        return func


class DateLTE(DateComparisonRule):
    """ Date less than or equal to comparison Rule """

    comparison_language = 'less than or equal to'

    def _evaluator(self):
        def func(x):
            try:
                return datetime.strptime(x, self.date_format) <= self.comparison_value
            except ValueError:
                return False

        return func


class DateLT(DateComparisonRule):
    """ Date less than comparison Rule """

    comparison_language = 'less than'

    def _evaluator(self):
        def func(x):
            try:
                return datetime.strptime(x, self.date_format) < self.comparison_value
            except ValueError:
                return False

        return func


class ColumnComparisonRule(Rule):
    """ Base column comparison Rule """

    _default_args = ('x',)

    def __init__(self, compare_to: Union[str, List[str]]):
        super().__init__()
        self.compare_to = compare_to

    def _prepare(self, data: Tuple[str, Dict]) -> tuple:
        return data[0], data[1][self.compare_to]


class OtherCantExist(ColumnComparisonRule):
    @staticmethod
    def _helper(val, other):
        if other not in ['', False]:
            if val not in ['', False]:
                return False
            else:
                return True
        else:
            return True

    def _evaluator(self):
        return lambda x, y: self._helper(x, y)

    def _explain(self) -> str:
        return f"cannot contain a value if '{self.compare_to}' is populated"


class OtherMustExist(ColumnComparisonRule):
    @staticmethod
    def _helper(val, other):
        if other in ['', False]:
            if val not in ['', False]:
                return False
            else:
                return True
        else:
            return True

    def _evaluator(self):
        return lambda x, y: self._helper(x, y)

    def _explain(self) -> str:
        return f"must contain a value if '{self.compare_to}' is populated"


class GreaterThanColumn(ColumnComparisonRule):
    """ Greater than compared column Rule """

    def _evaluator(self):
        return lambda x, y: x > y

    def _explain(self) -> str:
        return f"must be greater than column '{self.compare_to}'"


class GreaterThanOrEqualColumn(ColumnComparisonRule):
    """ Greater than compared column Rule """

    def _evaluator(self):
        return lambda x, y: x >= y

    def _explain(self) -> str:
        return f"must be greater than or equal to column '{self.compare_to}'"


class LessThanColumn(ColumnComparisonRule):
    """ Less than compared column Rule """

    def _evaluator(self):
        return lambda x, y: x < y

    def _explain(self) -> str:
        return f"must be less than column '{self.compare_to}'"


class LessThanOrEqualColumn(ColumnComparisonRule):
    """ Less than compared column Rule """

    def _evaluator(self):
        return lambda x, y: x <= y

    def _explain(self) -> str:
        return f"must be less than or equal to column '{self.compare_to}'"


class NotNullIfCompare(ColumnComparisonRule):

    def __init__(self, compare_to: [str, List]):
        super().__init__(compare_to=compare_to)

    def _prepare(self, data: Tuple[str, Dict]) -> tuple:
        return data

    def _null_ok(self, data):
        empty_val = data[0] in ['', False]
        empty_compare = True
        if isinstance(self.compare_to, str):
            empty_compare = data[1][self.compare_to] in ['', False]
        elif isinstance(self.compare_to, list):
            empty_compare = any([v for k, v in data[1].items() if k in self.compare_to]) in ['', False]
        if not empty_compare and empty_val:
            return False
        else:
            return True

    def _evaluator(self):
        return lambda x, y: self._null_ok((x, y))

    def _explain(self) -> str:
        compare_msg = ''
        if isinstance(self.compare_to, str):
            compare_msg = self.compare_to
        elif isinstance(self.compare_to, list):
            compare_msg = "' or '".join(self.compare_to)
        return f"cannot be blank if '{compare_msg}' contains a value"


class NotNullIfOtherEquals(NotNullIfCompare):
    """ Cell cannot be null if other has specified value(s) """
    _default_args = ('x', 'x')

    def __init__(self, compare_to: str, values: Union[str, List[str]]):
        self.values = [values] if isinstance(values, str) else values
        super().__init__(compare_to=compare_to)

    def _null_ok(self, data):
        empty_val = data[0] in ['', False]
        if empty_val and data[1][self.compare_to] in self.values:
            return False
        else:
            return True

    def _explain(self) -> str:
        return f"cannot be blank if '{self.compare_to}' contains value: {self.values}"



import webbrowser
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep
from typing import Tuple
from warnings import warn

from rumydata.table import Layout, CsvFile

__all__ = ['menu']

_format_options = {'markdown': 'md', 'html': 'html'}


def menu(layout: Layout) -> dict:
    """
    Layout menu

    This method takes a Layout class object as an argument and presents the user
    with a suite of options in a command line interface (CLI). The options
    include the generation technical documentation for the layout, or validation
    of a csv file, with output of the results.

    Outputs can be printed, saved, or opened (in a web-browser) in either
    markdown or html format, depending upon the user input.

    :param layout: a Layout object which will be used for the selected options.

    :return: a dictionary which describes the results of the interface selection
        choices. Not useful in the wild, but necessary for testing since the
        output is not as easy to verify.
    """
    options = {
        'View Documentation': (_documentation, dict(ext='html', output='open')),
        'Generate Documentation': (_documentation, {}),
        'View Validation': (_validation, dict(ext='html', output='open')),
        'Generate Validation': (_validation, {})
    }

    print("What to do with this layout?")
    choice = _select_option(options)[0]
    return choice[0](layout, **choice[1])


def _documentation(layout: Layout, ext: str = None, output: str = None) -> dict:
    """ Run Layout documentation through menu """
    return _doc_out(*_doc_gen(layout, ext), output)


def _validation(layout: Layout, ext: str = None, output: str = None) -> dict:
    """ Run validation result through menu """
    return _doc_out(*_file_check(layout, ext), output)


def _file_check(layout: Layout, ext) -> Tuple[str, str]:
    """ Perform a check of a csv file with (mostly) default params """
    if not ext:
        print("How to format the documentation?")
        ext = _select_option(_format_options)[0]

    if ext == 'html':
        try:  # check if markdown to html conversion is available
            __import__('markdown')
        except ModuleNotFoundError:
            warn('markdown module not available; falling back to raw md')
            ext = 'md'

    p = input('What is the file path to validate?\n > ')
    errors = CsvFile(layout).check(p, doc_type=ext)
    if errors:
        return errors, ext


def _doc_gen(layout: Layout, extension: str = None) -> Tuple[str, str]:
    """ Generate documentation from a layout """

    if not extension:
        print("How to format the documentation?")
        extension = _select_option(_format_options)[0]

    if extension == 'html':
        try:  # check if markdown to html conversion is available
            __import__('markdown')
        except ModuleNotFoundError:
            warn('markdown module not available; falling back to raw md')
            extension = 'md'

    return layout.documentation(doc_type=extension), extension


def _doc_out(document: str, extension: str = None, output: str = None):
    """ Output documentation or result """
    summary = dict(document=document, extension=extension)
    output_options = {'print': print, 'open': _open_doc, 'save': _save_doc}

    if output:
        choice = output_options[output]
    else:
        print("How to output the results?")
        choice, output = _select_option(output_options)

    summary['output'] = output

    output_kwargs = dict(ext=extension) if choice is _open_doc else {}
    choice(document, **output_kwargs)
    return summary


def _save_doc(doc: str):
    """ Save str to a specified text file path """
    ip = input('What path to save the file?\n > ')
    p = Path(ip)
    print(f"Writing to {p.absolute().as_posix()}")
    p.write_text(doc)


def _open_doc(doc: str, ext: str):
    """ Open str as a web-browser document """
    with TemporaryDirectory() as td:
        p = Path(td, f'documentation.{ext}')
        p.write_text(doc)
        if p.suffix == '.html':
            url = f'file:///{p.as_posix()}'
        else:
            url = p.as_posix()
        webbrowser.open(url)
        sleep(2)


def _select_option(options: dict):
    """ Present a dictionary of options to the user for selection """
    for ix, (k, v) in enumerate(options.items()):
        print(f'[{str(ix)}] {k}')
    choice = input("Choose an option by number or name: ")
    print('')

    try:
        choice = int(choice)  # treat the input as if it was an integer
        choice = list(options.keys())[choice]
    except ValueError:  # otherwise assume it was a key value
        pass

    return options[choice], choice



"""
data validation rules

This submodule contains the various rules that are used to validate the contents
of a file, to determine if it meets the specifications of a layout.

The rules are divided by the various types:
 - cell: the most basic rule type. A cell rule generally receives a single
   value, but can also perform comparison to other value that exist in the same
   row.
 - column: a more complex rule type which requires knowledge of the entire
   list of values present in a single column in order to determine validity.
 - row: rules that are applied to an entire row. These are not generally meant
   to be extended by users of this package.
 - header: rules that extend the concept of row rules, but are meant to apply
   specifically to the header row of the file.
 - file: rules that apply to the file, including whether the file exists, or
   matches a particular naming convention.
"""

from rumydata.rules import cell, column, row, header, table

__all__ = ['cell', 'column', 'row', 'header', 'table']
